#include "globals.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

// Prototypes
void CALLBACK_display(void);
void CALLBACK_keyboard(unsigned char key, int x, int y);
void CALLBACK_idle();
void CALLBACK_reshape(int w, int h);
void CALLBACK_display(void);
void CALLBACK_mouseclick (int mouse, int state, int x, int y);
void CALLBACK_mousemove (int x, int y);
void drawStrokeOval(STROKE s);
void drawStrokeRect(STROKE s);
void initDisplay(int argc, char** argv);
void renderPainting();
void LoadTextures();
void uploadTexture(IplImage* thecanvas, IplImage* alphaimg, int texnum);
void genstroke(int x, int y);
void regenerateBrushSizeImage();
bool checkIfInControl();
void setBrushSize(float newsize);


#define TEXTURE_IMAGE  (0)
#define TEXTURE_BRUSH  (1)
#define TEXTURE_CANVAS (2)
#define TEXTURE_BRUSHCONTROL (3)

#define USE_MIPMAP

// Globals
IplImage*	g_sourceImage=NULL;
IplImage*	g_edgeMagnitude=NULL;
IplImage*	g_edgeDirection=NULL;
bool		g_mouseDown=false;
PAINTING	g_painting;
GLuint		g_painting_displaylist;
vector<CvPoint2D32f> g_circle;
CvSize		g_viewport_size;
int			g_strokecount=0;
CvSize		g_originalSize;
float		g_originalAspect;
GLuint		g_textures[4];
CvPoint2D32f	g_lastclick=cvPoint2D32f(-1,-1);
float		g_jitterscale;
float		g_currentBrushSize;
IplImage*	g_brushimage;



int main (int argc, char** argv) {


	if (argc!=4) {
		printf("SYNTAX:  %s <source image> <brush texture> <canvas image>\n",argv[0]);
		exit(1);
	}

	// load source image
	IplImage* scratch=cvLoadImage(argv[1],1);
	if (!scratch) {
		printf("Could not open source image %s\n",argv[1]);
		exit(1);
	}

	// load canvas image
	IplImage* canvasscratch=cvLoadImage(argv[3],1);
	if (!canvasscratch) {
		printf("Could not open canvas image %s\n",argv[3]);
		exit(3);
	}


	// load brush
	IplImage* brushscratch=cvLoadImage(argv[2],1);
	if (!brushscratch) {
		printf("Could not open brush texture %s\n",argv[2]);
		exit(2);
	}
	IplImage* brush=cvCreateImage(cvSize(BRUSH_TEX_X,BRUSH_TEX_Y),8,3);
	cvResize(brushscratch,brush,CV_INTER_CUBIC);
	cvReleaseImage(&brushscratch);

	IplImage* canvas=cvCreateImage(cvSize(IMG_INTERNAL_RESIZE,IMG_INTERNAL_RESIZE),8,3);
	cvResize(canvasscratch,canvas,CV_INTER_CUBIC);
	cvReleaseImage(&canvasscratch);

	g_originalSize=cvGetSize(scratch);
	g_originalAspect=((float)g_originalSize.width/(float)g_originalSize.height);
	g_sourceImage=cvCreateImage(cvSize(IMG_INTERNAL_RESIZE,IMG_INTERNAL_RESIZE),8,3);
	cvResize(scratch,g_sourceImage,CV_INTER_CUBIC);	

	IplImage* dx=cvCreateImage(cvGetSize(scratch),32,1);
	IplImage* dy=cvCreateImage(cvGetSize(scratch),32,1);
	IplImage* g_edgeMagnitudeScratch=cvCreateImage(cvGetSize(scratch),32,1);
	IplImage* g_edgeDirectionScratch=cvCreateImage(cvGetSize(scratch),32,1);

	IplImage* scratchGrey8=cvCreateImage(cvGetSize(scratch),8,1);
	IplImage* scratchGrey32=cvCreateImage(cvGetSize(scratch),32,1);
	cvCvtColor(scratch,scratchGrey8,CV_BGR2GRAY);
	cvCvtScale(scratchGrey8,scratchGrey32,1.0f/255.0f,0);

	cvSmooth(scratchGrey32,scratchGrey32,CV_GAUSSIAN,11);

	cvSobel( scratchGrey32, dx, 1, 0, 3);
	cvSobel( scratchGrey32, dy, 0, 1, 3); 
	
	cvCartToPolar(dx,dy,g_edgeMagnitudeScratch,g_edgeDirectionScratch);

//	cvNamedWindow("mag",1);
//	cvShowImage("mag",g_edgeMagnitudeScratch);

	cvReleaseImage(&scratchGrey8);
	cvReleaseImage(&scratchGrey32);
	cvReleaseImage(&dx);
	cvReleaseImage(&dy);

	g_edgeMagnitude=cvCreateImage(cvSize(IMG_INTERNAL_RESIZE,IMG_INTERNAL_RESIZE),32,1);
	g_edgeDirection=cvCreateImage(cvSize(IMG_INTERNAL_RESIZE,IMG_INTERNAL_RESIZE),32,1);

	cvResize(g_edgeMagnitudeScratch,g_edgeMagnitude,CV_INTER_LINEAR);
	cvResize(g_edgeDirectionScratch,g_edgeDirection,CV_INTER_LINEAR);

	cvReleaseImage(&g_edgeMagnitudeScratch);
	cvReleaseImage(&g_edgeDirectionScratch);
	cvReleaseImage(&scratch);	

   glutInit(&argc, argv);
   glutInitDisplayMode (GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH | GLUT_ALPHA);
   glutInitWindowSize (800.0f, 800.0f/g_originalAspect);
   glutCreateWindow ("University of Surrey/CVSSP - Impressionist Painting Demo");

   glutReshapeFunc (CALLBACK_reshape);
   glutKeyboardFunc (CALLBACK_keyboard);
   glutDisplayFunc (CALLBACK_display);
   glutIdleFunc (CALLBACK_idle);
   glutMouseFunc (CALLBACK_mouseclick);
   glutPassiveMotionFunc (CALLBACK_mousemove);
   glutMotionFunc (CALLBACK_mousemove);
   initDisplay(argc, argv);

   glutSetCursor(GLUT_CURSOR_CROSSHAIR);

   // Init textures
   glGenTextures(4,g_textures);

	// upload textures
   IplImage* mask=cvCreateImage(cvGetSize(g_sourceImage),8,1);
   cvSet(mask,cvScalarAll(255));
   uploadTexture(g_sourceImage,mask,TEXTURE_IMAGE);
   uploadTexture(canvas,mask,TEXTURE_CANVAS);
   cvReleaseImage(&mask);

   mask=cvCreateImage(cvGetSize(brush),8,3);
   IplImage* bmask=cvCreateImage(cvGetSize(brush),8,1);
   cvCvtColor(brush,bmask,CV_BGR2GRAY);
   cvConvertScale(bmask,bmask,STROKE_TRANSP,0);
   cvSet(mask,cvScalarAll(255));
   uploadTexture(mask,bmask,TEXTURE_BRUSH);
   cvReleaseImage(&mask);
   cvReleaseImage(&bmask);

   // setup brush image
   g_brushimage=cvCreateImage(cvSize(128,256),8,3);
   setBrushSize(0.9);
   regenerateBrushSizeImage();

   glutMainLoop();

}


void uploadTexture(IplImage* thecanvas, IplImage* alphaimg, int texnum) {

	char* uploadData;
	int i;
	
	uploadData=(char*)calloc((thecanvas->width+2)*(thecanvas->height+2)*4,sizeof(char));	

	for (i=0; i<thecanvas->width*thecanvas->height; i++) {
		*(uploadData+(i*4)+0)=*(thecanvas->imageData+(i*3)+2);
		*(uploadData+(i*4)+1)=*(thecanvas->imageData+(i*3)+1);
		*(uploadData+(i*4)+2)=*(thecanvas->imageData+(i*3)+0);
		*(uploadData+(i*4)+3)=*(alphaimg->imageData+i);
	}


	glBindTexture(GL_TEXTURE_2D,g_textures[texnum]);
	
	glPixelStorei (GL_UNPACK_ALIGNMENT,1);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP); 
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

#ifdef USE_MIPMAP
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_LINEAR);
#endif

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	gluBuild2DMipmaps (GL_TEXTURE_2D, 4, thecanvas->width, thecanvas->height, GL_RGBA, GL_UNSIGNED_BYTE, uploadData);

	free(uploadData);

}

// Create output window
void initDisplay(int argc, char** argv) {


   glClearColor (0.0, 0.0, 0.0, 0.0);

   /* Material */
   glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);

   glEnable(GL_COLOR_MATERIAL);

   /* Lighting */
   glDisable(GL_LIGHTING); // flat shading is fine

        /* OpenGL Rendering engine parameters */
   glShadeModel (GL_FLAT);
   glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);

   g_painting_displaylist = glGenLists(1);

   // Pre-compute vertices of unit circle
   for (float th=0; th<2*PI; th+=(2*PI)/CIRCLE_INTERVAL) {
	   g_circle.push_back(cvPoint2D32f(cos(th),sin(th)));
   }
   g_circle.push_back(cvPoint2D32f(1.0f,0.0f));


}

void CALLBACK_keyboard(unsigned char key, int x, int y) {
   switch (key) {
          case 0x1b:
                  exit (0);
                  break;
   }

}

void CALLBACK_idle() {

}

void CALLBACK_mouseclick (int mouse, int state, int x, int y) {

	g_mouseDown=(state==GLUT_DOWN);
	CALLBACK_mousemove(x,y);
	
}

void CALLBACK_reshape(int w, int h) {

   glViewport(0, 0, w, h);

   g_viewport_size=cvSize(w,h);

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   glOrtho(0,1,0,1,-1,1);


}

void CALLBACK_display(void) {

        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glDisable( GL_DEPTH_TEST );
    glEnable( GL_BLEND );
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

		// Canvas
		glColor3f(1.0f,1.0f,1.0f);
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D,g_textures[TEXTURE_CANVAS]);

		glBegin(GL_QUADS);
			glNormal3f(0,0,1);
			glTexCoord2f(0,1);
			glVertex3f(0,0,BG_IMAGE_DEPTH);
			glTexCoord2f(1,1);
			glVertex3f(1,0,BG_IMAGE_DEPTH);
			glTexCoord2f(1,0);
			glVertex3f(1,1,BG_IMAGE_DEPTH);
			glTexCoord2f(0,0);
			glVertex3f(0,1,BG_IMAGE_DEPTH);
		glEnd();

		// strokes
		glBindTexture(GL_TEXTURE_2D,g_textures[TEXTURE_BRUSH]);

		renderPainting();

		// Draw mini-image overlay
		glColor3f(1.0f,1.0f,1.0f);
		glBindTexture(GL_TEXTURE_2D,g_textures[TEXTURE_IMAGE]);
		glBegin(GL_QUADS);
			glNormal3f(0,0,1);
			glTexCoord2f(0,1);
			glVertex3f(1.0f-INSET_IMAGE_SIZE,1.0f-INSET_IMAGE_SIZE,INSET_IMAGE_DEPTH);
			glTexCoord2f(1,1);
			glVertex3f(1,1.0f-INSET_IMAGE_SIZE,INSET_IMAGE_DEPTH);
			glTexCoord2f(1,0);
			glVertex3f(1,1,INSET_IMAGE_DEPTH);
			glTexCoord2f(0,0);
			glVertex3f(1.0f-INSET_IMAGE_SIZE,1,INSET_IMAGE_DEPTH);
		glEnd();

		
		glDisable(GL_TEXTURE_2D);

		glColor4f(1.0f,1.0f,0.0f,1.0f);
		glLineWidth(3.0f);
		glBegin(GL_LINES);
			glVertex3f(1.0f-INSET_IMAGE_SIZE,1.0f-INSET_IMAGE_SIZE,INSET_IMAGE_DEPTH-0.01);
			glVertex3f(1,1.0f-INSET_IMAGE_SIZE,INSET_IMAGE_DEPTH-0.01);
			glVertex3f(1,1,INSET_IMAGE_DEPTH-0.01);
			glVertex3f(1.0f-INSET_IMAGE_SIZE,1,INSET_IMAGE_DEPTH-0.01);
		glEnd();

		glColor4f(1.0f,1.0f,1.0f,1.0f);
		float xloc=(1.0f-INSET_IMAGE_SIZE)+(INSET_IMAGE_SIZE*g_lastclick.x);
		float yloc=1.0f-(INSET_IMAGE_SIZE*g_lastclick.y);
		glPushMatrix();
		glTranslatef(xloc,yloc,0);
		glutSolidSphere(0.005,4,4);
		glPopMatrix();

		// Draw brush control
		glEnable(GL_TEXTURE_2D);
		glColor3f(1.0f,1.0f,1.0f);
		glBindTexture(GL_TEXTURE_2D,g_textures[TEXTURE_BRUSHCONTROL]);
		glBegin(GL_QUADS);
			glNormal3f(0,0,1);
			glTexCoord2f(0,1);
			glVertex3f(1.0f-CONTROL_IMAGE_SIZE,CONTROL_IMAGE_SIZE*4,CONTROL_IMAGE_DEPTH);
			glTexCoord2f(1,1);
			glVertex3f(1,CONTROL_IMAGE_SIZE*4,CONTROL_IMAGE_DEPTH);
			glTexCoord2f(1,0);
			glVertex3f(1,0,CONTROL_IMAGE_DEPTH);
			glTexCoord2f(0,0);
			glVertex3f(1.0f-CONTROL_IMAGE_SIZE,0,CONTROL_IMAGE_DEPTH);
		glEnd();


		// Flush / swap buffers
		glFlush();
        glutSwapBuffers();

}

void CALLBACK_mousemove (int x, int y){

	g_lastclick=cvPoint2D32f(((float)x/(float)g_viewport_size.width),((float)y/(float)g_viewport_size.height));

	if (g_mouseDown) {
		
		if (!checkIfInControl())
			genstroke(x,y);

		glutPostRedisplay();
	}

}

bool checkIfInControl() {

		// check if in the brush control
		if (g_lastclick.x>1.0f-CONTROL_IMAGE_SIZE && g_lastclick.y>(1.0f-(CONTROL_IMAGE_SIZE*4))) {
			// update brush size
			float splitdisplay=0.75*g_brushimage->height;
			float imgupper=(1.0f-(CONTROL_IMAGE_SIZE*4.0f));
			float imglower=1.0f;
		    float imgrange=imglower-imgupper;

			float slide_lower=1.0f-(imgrange*TRI_INSET);
			float slide_upper=imgupper+ (imgrange*(0.25+TRI_INSET));
//			printf("iu=%f il=%f su=%f sl=%f p=%f\n",imgupper,imglower,slide_upper,slide_lower,g_lastclick.y);

			if (g_lastclick.y>slide_upper && g_lastclick.y<slide_lower) {
				// its in the slider vertical zone!
				setBrushSize((g_lastclick.y-slide_upper)/(slide_lower-slide_upper));
				
				regenerateBrushSizeImage();
			}

			return true;
		}

		return false;

}

void setBrushSize(float newsize) {

	g_currentBrushSize=newsize;
	g_jitterscale=newsize*0.2;

}


void genstroke(int x, int y) {

		COLOUR c;
		int imgx=(int)(((float)x/(float)g_viewport_size.width)*g_sourceImage->width);
		int imgy=(int)(((float)y/(float)g_viewport_size.height)*g_sourceImage->height);
		imgx=max(0,min(imgx,g_sourceImage->width-1));
		imgy=max(0,min(imgy,g_sourceImage->height-1));
		float img_mag=*(((float*)g_edgeMagnitude->imageData)+imgx+imgy*g_edgeMagnitude->width);
		float img_dir=*(((float*)g_edgeDirection->imageData)+imgx+imgy*g_edgeDirection->width);
		unsigned char* ptr=((unsigned char*)g_sourceImage->imageData)+imgx*3+imgy*g_sourceImage->widthStep;
		c.red=(float)ptr[2]/255.0f;
		c.green=(float)ptr[1]/255.0f;
		c.blue=(float)ptr[0]/255.0f;

		c.red+=g_jitterscale*(((float)rand()/(float)RAND_MAX)-((float)rand()/(float)RAND_MAX));
		c.green+=g_jitterscale*(((float)rand()/(float)RAND_MAX)-((float)rand()/(float)RAND_MAX));
		c.blue+=g_jitterscale*(((float)rand()/(float)RAND_MAX)-((float)rand()/(float)RAND_MAX));
		if (c.red>1.0f)
			c.red=1.0f;
		if (c.green>1.0f)
			c.green=1.0f;
		if (c.blue>1.0f)
			c.blue=1.0f;
		if (c.red<0)
			c.red=0;
		if (c.green<0)
			c.green=0;
		if (c.blue<0)
			c.blue=0;
		// add a stroke
		STROKE s;
		s.x=(float)x/(float)g_viewport_size.width;
		s.y=1.0f-((float)y/(float)g_viewport_size.height);
		s.order=(float)(g_strokecount++)/(float)10000;
		s.scale=(g_currentBrushSize*(MAX_BRUSH_SIZE-MIN_BRUSH_SIZE))+MIN_BRUSH_SIZE;
//		printf("mag=%f dir=%f\n",img_mag,img_dir);
		s.aspect=0.8f+(img_mag*3.0f);//0.5f;
		s.orient=(PI/2)-img_dir;
		s.colour=c;
		g_painting.strokelist.push_back(s);

}


void regenerateBrushSizeImage() {

	cvSet(g_brushimage,cvScalarAll(0xf0)); // white out image
	CvPoint tri[3];
	float splitdisplay=0.75*g_brushimage->height;
	cvLine(g_brushimage,cvPoint(0,splitdisplay),cvPoint(g_brushimage->width-1,splitdisplay),cvScalarAll(0),2);

	float inset=TRI_INSET;
	tri[0].y=tri[1].y=inset*g_brushimage->height;
	tri[0].x=inset*g_brushimage->width;
	tri[1].x=(1.0f-inset)*g_brushimage->width;
	tri[2].y=(1.0f-inset)*splitdisplay;
	tri[2].x=0.5*g_brushimage->width;

	cvFillConvexPoly(g_brushimage,tri,3,cvScalarAll(0x40));
	float barpos=((tri[2].y-tri[0].y)*(1.0f-g_currentBrushSize))+tri[0].y;
	cvRectangle(g_brushimage,cvPoint(tri[0].x+1,barpos-3),cvPoint(tri[1].x-1,barpos+3),cvScalar(0,255,0,0),-1);

	float rad=((g_brushimage->height-splitdisplay)*0.48)*(0.2+pow((double)g_currentBrushSize,0.7)*0.8);
	cvCircle(g_brushimage,cvPoint(g_brushimage->width*0.5,splitdisplay+(g_brushimage->height-splitdisplay)*0.5),rad,cvScalarAll(0),-1);

	IplImage* mask=cvCreateImage(cvGetSize(g_brushimage),8,1);
	cvSet(mask,cvScalarAll(255));
	uploadTexture(g_brushimage,mask,TEXTURE_BRUSHCONTROL);
	cvReleaseImage(&mask);

}


void renderPainting() {

	int sc=g_painting.strokelist.size();
	int base=(sc/RESEED_DISPLAYLIST_AT)*RESEED_DISPLAYLIST_AT;
	int residual=sc%RESEED_DISPLAYLIST_AT;

	vector<STROKE>::iterator sit;

	
	sit=g_painting.strokelist.begin();
	if (!residual) {
		// Regenerate display list
		glDeleteLists(g_painting_displaylist,1);
		glNewList(g_painting_displaylist, GL_COMPILE_AND_EXECUTE);
		for (int i=0; i<base; i++) {
			drawStrokeRect(*(sit++));
		}
		glEndList();
	}
	else {
		// Run the existing display list
		sit+=base;
		glCallList(g_painting_displaylist);
	}

	
	// Draw any residual strokes
	while (sit!=g_painting.strokelist.end()) {
		drawStrokeRect(*(sit++));
	}

}

void drawStrokeOval(STROKE s) {

	float asp=s.aspect*g_originalAspect;

	glColor3f(s.colour.red,s.colour.green,s.colour.blue);
	glBegin(GL_TRIANGLE_FAN);
	for (vector<CvPoint2D32f>::iterator pt=g_circle.begin(); pt!=g_circle.end(); pt++) {
		float xscaled=pt->x*s.scale;//*asp;
		float yscaled=pt->y*(s.scale/asp);
		glVertex3f(s.x+
				cos(s.orient)*xscaled-sin(s.orient)*yscaled,
				s.y+
				sin(s.orient)*xscaled+cos(s.orient)*yscaled,
				s.order);
	}
	glEnd();

}

void drawStrokeRect(STROKE s) {

	float asp=s.aspect*g_originalAspect;
	float xscaled=s.scale;//*asp;
	float yscaled=(s.scale/asp)*0.5;

	glColor3f(s.colour.red,s.colour.green,s.colour.blue);
	glBegin(GL_QUADS);
		glNormal3f(0,0,1);
		glTexCoord2f(0,1);
		double x=-xscaled; double y=-yscaled;
		glVertex3f(s.x+x*cos(s.orient)-y*sin(s.orient),s.y+x*sin(s.orient)+y*cos(s.orient),s.order);
		glTexCoord2f(1,1);
		x=xscaled; y=-yscaled;
		glVertex3f(s.x+x*cos(s.orient)-y*sin(s.orient),s.y+x*sin(s.orient)+y*cos(s.orient),s.order);
		x=xscaled; y=yscaled;
		glTexCoord2f(1,0);
		glVertex3f(s.x+x*cos(s.orient)-y*sin(s.orient),s.y+x*sin(s.orient)+y*cos(s.orient),s.order);
		x=-xscaled; y=yscaled;
		glTexCoord2f(0,0);
		glVertex3f(s.x+x*cos(s.orient)-y*sin(s.orient),s.y+x*sin(s.orient)+y*cos(s.orient),s.order);
	glEnd();

}